# © Canonical 2015
# Licensed under AGPLv3

import json
import logging
import os
import random
import subprocess
import sys
import urllib.parse

from configparser import ConfigParser

import requests
import requests_oauthlib

# Superset of Agent conf, TODO: should probably just take separate confs
# and merge them
DEFAULT_CONFIG = {
    'server': 'http://message-hub.spi.canonical.com/',
    'specmgr_url': 'http://spec-manager.spi.canonical.com/',
    'results_url': 'http://result-enumerator.spi.canonical.com/orgs/%(organization_id)s/results/',
    'id': 'e2e-id',
    'group': 'e2e-group',
    'interval': 10,
    'provcommand': "/bin/echo provisioning now",
    'setupcommand': "/bin/echo setting up now",
    'testcommand': "/bin/echo -e '{\"foo\": \"bar\"}' > spi_test_result.json \\\c",
    'logging_basedir': os.path.join(os.path.dirname(__file__), 'logs'),
    'organization_id': 'e2e-org',
    'platform': 'amd64',
    'image_name': 'e2e-img',
    'spec_name': 'e2e-spec',
    'image_unique_id': 'DEADBEEF',
    'test_payload': {"foo": "bar"},
    'network_timeout': 10,
}

RST = "\033[0m"
BLD = "\033[1m"
GRY = "\033[1;30m"
RED = "\033[1;31m"
GRN = "\033[1;32m"
YEL = "\033[1;33m"

logging.basicConfig(format='%(asctime)s %(levelname)-5.5s %(message)s')
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

if 'BuiltinImporter' in str(__loader__):
    nuitka = True
else:
    nuitka = False


class E2e(object):

    def __init__(self, arguments):
        # Parse CLI
        self.config = ConfigParser()
        self.confpath = arguments['<configfile>']
        self.interactive = arguments['-i']
        self.exit_code = 0

        if arguments['--default-config']:
            self.config['config'] = DEFAULT_CONFIG
            self.config.write(sys.stdout)
            exit(0)

        # Load config
        self.config['DEFAULT'] = DEFAULT_CONFIG
        if self.confpath is not None:
            self.config.read(self.confpath)
        if 'config' not in self.config.sections():
            self.config['config'] = {}

        self.id = self.config.get('config', 'id')
        self.group = self.config.get('config', 'group') or None
        self.organization_id = self.config.get('config', 'organization_id')
        self.platform = self.config.get('config', 'platform')
        self.interval = self.config.getfloat('config', 'interval')
        self.server = self.config.get('config', 'server')
        self.specmgr_url = self.config.get('config', 'specmgr_url')
        self.results_url = self.config.get('config', 'results_url')
        self.image_name = self.config.get('config', 'image_name')
        self.test_payload = self.config.get('config', 'test_payload')
        self.timeout = self.config.getint('config', 'network_timeout')
        try:
            # if loaded from .ini, this will be a string that needs to be parsed
            self.test_payload = json.loads(self.test_payload)
        except:
            pass
        rand_token = str(random.randint(10000000, 99999999))
        self.spec_name = "{}_{}".format(
            self.config.get('config', 'spec_name'), rand_token)
        self.image_unique_id = "{}_{}".format(
            self.config.get('config', 'image_unique_id'), rand_token)

        self.manifest_url = urllib.parse.urljoin(
            self.specmgr_url,
            '/'.join(['orgs', self.organization_id, 'products/manifests']))
        self.spec_url = urllib.parse.urljoin(
            self.specmgr_url,
            '/'.join(['orgs', self.organization_id, 'tests/specs']))
        self.image_url = urllib.parse.urljoin(
            self.specmgr_url,
            '/'.join(['orgs', self.organization_id, 'images']))
        self.snap_revision_url = urllib.parse.urljoin(
            self.specmgr_url,
            '/'.join(['orgs', self.organization_id, 'snap-revisions']))

        if not os.path.isfile(self.confpath):
            logger.error("Can't read config file: %s", self.confpath)
            exit(1)

        # prepare a normal session or an authenticated one if have the config
        if 'auth' in self.config:  # Has authentication configured
            logger.info("Working in Authenticated mode")
            self.session = requests_oauthlib.OAuth1Session(
                self.config.get('auth', 'consumer_key'),
                client_secret=self.config.get('auth', 'consumer_secret'),
                resource_owner_key=self.config.get('auth', 'token_key'),
                resource_owner_secret=self.config.get('auth', 'token_secret'),
            )
        else:
            logger.warning("Authentication configuration not found")
            self.session = requests.Session()

    def pause(self):
        if self.interactive:
            input('--- press enter ---')

    def annotate(self, msg, color=GRN):
        if self.interactive:
            logger.info(color + msg + RST)
        else:
            logger.info(msg)

    def prompt(self, msg):
        self.annotate(msg)
        self.pause()

    def cleanup_specs(self):
        """Remove all existing specs."""
        spec_list = self.session.get(self.spec_url, timeout=self.timeout).json()['specs']
        for spec in spec_list:
            self.delete_spec(spec['id'], verbose=False)

    def create_spec(self):
        d = dict({
            'organization_id': self.organization_id,
            'platform': self.platform,
            'image_name': self.image_name,
            'name': self.spec_name,
            'test_payload': self.test_payload,
        })
        logger.info('POST %s: %s', self.spec_url, json.dumps(d, indent=2))
        self.pause()

        res = self.session.post(self.spec_url, json=d)
        if res.status_code == 409:
            logger.error('Spec exists, ignore this e2e run')
            exit(0)
        elif res.status_code not in (200, 201, 202) or res.json() is None:
            logger.error(
                'Error creating spec: %r, %r',
                res.status_code,
                res.text)
            raise Exception

        js = res.json()
        self.spec_id = js['spec']['id']
        self.annotate('Created Spec id={}, result:'.format(self.spec_id))
        logger.info('Location: %s', res.headers['Location'])
        logger.info(json.dumps(js, indent=2))
        self.pause()

    def delete_spec(self, spec_id=None, verbose=True):
        if spec_id is None:
            spec_id = self.spec_id
        url = '{}/{}'.format(self.spec_url, spec_id)
        if verbose:
            self.annotate('Cleaning up, delete Test Spec')
            self.annotate('Note: normally Test Specs are long-lived', color=GRY)
            logger.info('DELETE ' + url)
            self.pause()

        res = self.session.delete(url, timeout=self.timeout)
        if res.status_code not in (200, 202) or res.json() is None:
            logger.error(
                'Error deleting spec: %r, %r',
                res.status_code,
                res.text)
            raise Exception

        if verbose:
            self.annotate('Deleted Spec:')
            logger.info(res.json())

    def submit_image(self):
        d = dict({
            'image_reference': 'http://foo.com/bar',
            'image_unique_id': self.image_unique_id,
            'image_name': self.image_name})
        logger.info('POST %s:%s', self.image_url, json.dumps(d, indent=2))
        self.pause()

        res = self.session.post(self.image_url, json=d, timeout=self.timeout)
        if res.status_code not in (200, 201, 202) or res.json() is None:
            logger.error(
                'Error submitting image: %r, %r',
                res.status_code,
                res.text)
            raise Exception

        self.annotate('Submitted image, result (Test Opportunity IDs):')
        logger.info(json.dumps(res.json(), indent=2))
        self.testopp_ids = set(res.json()['ids'])
        self.pause()

    def cleanup_manifests(self):
        """Deactivate all existing manifests."""
        manifest_list = self.session.get(
            self.manifest_url, timeout=self.timeout).json()['manifests']
        for manifest in manifest_list:
            self.delete_manifest(manifest['id'])

    def create_manifest(self):
        self.annotate('As Product Manager, Create Manifest via Spec Manager')
        d = {
            'organization_id': self.organization_id,
            'image_name': self.image_name,
            'release': '15.04',
            'base_image_reference': 'http://foo.com/bar',
            'platform': self.platform,
            'name': self.image_name,
            'snaps': [
                ['snap-a', 1, True],
                ['snap-b', None, False]]}
        logger.info('POST %s: %s', self.manifest_url, json.dumps(d, indent=2))
        self.pause()

        res = self.session.post(self.manifest_url, json=d, timeout=self.timeout)
        if res.status_code == 409:
            logger.warn(
                'Manifest exists, probably not cleaned up '
                'from a failed e2e run')
            # XXX it would be nice if the specmgr returned the existing entity
            raise Exception
        elif res.status_code not in (200, 201, 202) or res.json() is None:
            logger.error(
                'Error creating manifest: %r, %r',
                res.status_code,
                res.text)
            raise Exception

        js = res.json()
        self.manifest_id = js['manifest']['id']
        self.annotate(
            'Created Manifest id={}, result:'.format(
                self.manifest_id))
        logger.info('Location: %s', res.headers['Location'])
        logger.info(json.dumps(js, indent=2))
        self.pause()

    def delete_manifest(self, manifest_id=None):
        if manifest_id is None:
            manifest_id = self.manifest_id
        self.annotate('Cleaning up, delete Manifest')
        self.annotate('Note: normally Manifests are long-lived', color=GRY)
        url = '{}/{}'.format(self.manifest_url, manifest_id)
        logger.info('DELETE ' + url)
        self.pause()
        res = self.session.delete(url, timeout=self.timeout)
        if res.status_code not in (200, 202) or res.json() is None:
            logger.error(
                'Error deleting manifest: %r, %r',
                res.status_code,
                res.text)
            raise Exception
        self.annotate('Deleted Manifest:')
        logger.info(res.json())

    def submit_snap_revision(self):
        self.annotate(
            'As Developer hacking code, Submit a new revision '
            'of a snap to Spec Manager')
        d = {
            'name': 'snap-b',
            'revision': 4
        }
        logger.info(
            'POST %s:%s',
            self.snap_revision_url,
            json.dumps(
                d,
                indent=2))
        self.pause()

        res = self.session.post(self.snap_revision_url, json=d, timeout=self.timeout)
        if res.status_code not in (200, 201, 202) or res.json() is None:
            logger.error(
                'Error submitting snap revision: %r, %r',
                res.status_code,
                res.text)
            raise Exception

        self.annotate('Submitted snap, result (Test Opportunity IDs):')
        logger.info(json.dumps(res.json(), indent=2))
        self.testopp_ids = set()
        self.image_unique_ids = set()
        for d in res.json()['ids']:
            self.testopp_ids.add(d['test_opportunity_id'])
            self.image_unique_ids.add(d['image_unique_id'])
        self.pause()

    def run_agent(self, reg_only=False):
        try:
            if reg_only:
                if nuitka:
                    subprocess.check_call(
                        './agent.py -v --register {}'.format(self.confpath),
                        shell=True)
                else:
                    subprocess.check_call(
                        '{} agent.py -v --register {}'.format(sys.executable, self.confpath),
                        shell=True)
            else:
                if nuitka:
                    subprocess.check_call(
                        './agent.py -v --debug --drain {}'.format(self.confpath),
                        shell=True)
                else:
                    subprocess.check_call(
                        '{} agent.py -v --debug --drain {}'.format(sys.executable, self.confpath),
                        shell=True)
        except Exception as e:
            logger.fatal('Exception calling Agent: %r', e)
            raise

    def query_results(self, **kw):
        """Make a query and return the results."""
        res = self.session.get(self.results_url.rstrip('/'), params=kw, timeout=self.timeout)
        logger.info('GET %s', res.url)
        if res.status_code != 200 or res.json() is None:
            logger.error('Error getting results: %r, %r', res.status_code, res.text)
            raise Exception
        logger.info(json.dumps(res.json(), indent=2))
        return res.json()

    def check_multiple_results(self):
        """Like check_result but checks multiple image_unique_ids instead of one.

        This is needed for the snap revision workflow, where we can get test
        opps accross multiple manifests.
        """

        result_testopp_ids = set()
        for image_unique_id in self.image_unique_ids:
            res = self.query_results(image_unique_id=image_unique_id)
            result_testopp_ids = result_testopp_ids.union(
                set([r['id'] for r in res['test_results']]))
        if self.testopp_ids - result_testopp_ids:
            logger.error('One of our tests is missing from the results')
            self.exit_code = 1
        else:
            self.annotate('All our tests have produced results.')
        self.pause()

    def check_results(self):
        """Check that our test opp generated a result."""
        results = self.query_results(spec_name=self.spec_name)
        result_testopp_ids = set([r['id'] for r in results['test_results']])
        if self.testopp_ids - result_testopp_ids:
            logger.error('One of our tests is missing from the results')
            self.exit_code = 1
        else:
            self.annotate('All our tests have produced results.')
        self.pause()

    def run(self):
        self.annotate('Override run in your own script')
