#!/usr/bin/env python3
"""
SPI Tutorial

Usage:
  tutorial.py [-i] [<configfile>]
  tutorial.py --default-config

Options:
  -h --help             Show this screen
  --default-config      Show default configuration
  --server              MessageHub URL override
  --specmgr             SpecManager URL override
  -i                    Interactive
"""

import sys
import time

try:
    import docopt
except:
    print('This script must be run in the virtualenv, via:')
    print('. env/bin/activate')
    print('./scripts/tutorial.py')
    sys.exit(0)

from e2e import E2e, logger,  GRY


class E2eImage(E2e):

    def run(self):
        self.cleanup_specs()

        self.annotate('Welcome to the SPI Tutorial!')
        self.annotate('----------------------------')

        self.annotate('We are going to walkthrough an image-based workflow,')
        self.annotate('performing a simple test case.')
        self.annotate('For the purposes of this tutorial we will skip any')
        self.annotate('actual device provisioning.')
        self.pause()

        self.annotate('First, we need to authenticate into the system.')
        self.annotate('You must create an account at https://login.ubuntu.com')
        self.annotate('Once your account is created, you run this comand:')
        self.annotate('agent.py --interactive-login')
        self.pause()

        self.annotate('That command produces some data you need to keep safe:')
        self.annotate('''
$ ./agent.py --interactive-login
Enter email address: some-email-account@canonical.com
Enter password:
Add this to your configuration file:


[auth]
consumer_key = RLfY3Lf
consumer_secret = KUA8gXZBZbcXO7ozTlcTEXo9KywjyTe93fN5JisSHvqDDHQuI7
token_key = hIXQiWdsQmggpYkPsjXPdyUMXxfi13TMH975wZ3jLcwOieGAhNvmKmKDkJiK
token_secret = l1LqOgHAQUot7JN8cDXRbEv3UFmnQg51h8IMyYyGET49NORJ0OWwYIrNct6Qz5Llq3s731
''')
        self.pause()
        self.annotate('NOTE: those are testing credentials, do not use them for real data.')
        self.pause()

        self.annotate('Specially important is the consumer_key, that has to be a member of your organization.')  # NOQA
        self.annotate('Usually, now you would have to request for "RLfY3Lf" to ')
        self.annotate('be added to the "tutorial-org" organization, but we have')
        self.annotate('done that for you already.')
        self.annotate('The organization_id has to be consistent all accross your usage of SPI.')
        self.pause()

        self.annotate('Now, we\'ll put on our QA hat and create a TestSpec')
        self.annotate('that describes Tests to be run for a given "image_name".')
        self.pause()
        self.annotate('The "test_payload" field is the meat of the TestSpec and will be available')
        self.annotate('to the setup and test stages of the Lab\'s Provisioning Kit')
        self.pause()
        self.annotate('SPI treats test_payload as arbitrary data.')
        self.annotate('It could be the URL of a tests tarball, JSON describing test suites, etc.')
        self.annotate('as long as the Test Spec author and the Provisioning Kit scripts')
        self.annotate('agree on the format.')
        self.pause()
        self.annotate('Here\'s an example Spec creation:')
        self.create_spec()

        self.annotate('Now we will setup a simple "Lab", running an Agent locally.')
        self.annotate('We are using the following agent config:')
        # XXX dump the raw file instead?
        logger.info({k: v for (k, v) in self.config.items('config')})

        self.prompt('We run the agent once to ensure we can register with SPI')
        self.annotate('./agent.py --register scripts/tutorial.ini')
        self.run_agent(reg_only=True)
        self.pause()
        self.annotate('At this point we have our Tests configured and ')
        self.annotate('an Agent ready to accept Test Opportunities.')
        self.annotate('The next step is to advertise an Image to trigger those tests.')
        self.pause()
        self.annotate('We use the same "image_name" we used before')
        self.annotate('so that the Test Spec we created will be matched.')
        self.pause()
        self.annotate('The "image_reference" is what will be handed over')
        self.annotate('to the Provisioning Kit "provcommand" in the Lab.')
        self.annotate('SPI treats it as arbitrary data, but typically this will be either')
        self.annotate('a URL to an image binary or other info for the provcommand to')
        self.annotate('install the image and ready the device for test run.')
        self.pause()
        self.annotate('"image_unique_id" is simply a hash or other identifier')
        self.annotate('unique to this submission')
        self.submit_image()
        self.annotate('Sleeping 1s to wait for the Test Opportunities to')
        self.annotate('propagate through the system')
        time.sleep(1)
        self.annotate('Now, back as the Lab Owner, we can run the agent')
        self.annotate('which should pick up at least 1 Test Opportunity, provision and run tests')
        self.annotate('Note this can also run continously in daemon mode', color=GRY)
        self.pause()
        self.run_agent()
        self.annotate('Completed agent run.')
        self.annotate('The Agent did its work and dispatched a Test Result to the SPI system.')

        self.annotate('Any participant of the organization can now query those results.')
        self.pause()
        self.check_results()
        self.annotate('Finally we\'ll cleanup, which would not be required in normal usage.')
        self.delete_spec(verbose=False)


if __name__ == '__main__':
    arguments = docopt.docopt(__doc__, version='SPI Tutorial v1')
    arguments['-i'] = True
    arguments['<configfile>'] = 'scripts/tutorial.ini'
    e2e = E2eImage(arguments)
    e2e.run()
    exit(e2e.exit_code)
